from __future__ import annotations

import logging
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    try:
        from treescope import (  # pyright: ignore[reportMissingImports]
            renderers,
            rendering_parts,
        )
    except ImportError:
        pass


def render_object_constructor(
    object_type: type[Any],
    attributes: Mapping[str, Any],
    path: str | None,
    subtree_renderer: "renderers.TreescopeSubtreeRenderer",
    roundtrippable: bool = False,
    color: str | None = None,
) -> "rendering_parts.Rendering":
    try:
        from treescope import rendering_parts  # pyright: ignore[reportMissingImports]
    except ImportError:
        logging.exception("Failed to import treescope.rendering_parts")
        raise

    if roundtrippable:
        constructor = rendering_parts.siblings(
            rendering_parts.maybe_qualified_type_name(object_type), "("
        )
        closing_suffix = rendering_parts.text(")")
    else:
        constructor = rendering_parts.siblings(
            rendering_parts.roundtrip_condition(roundtrip=rendering_parts.text("<")),
            rendering_parts.maybe_qualified_type_name(object_type),
            "(",
        )
        closing_suffix = rendering_parts.siblings(
            ")",
            rendering_parts.roundtrip_condition(roundtrip=rendering_parts.text(">")),
        )

    # Sort attributes so attributes with default values are at the end.
    attributes_list = [
        (name, value, is_default) for name, (value, is_default) in attributes.items()
    ]
    attributes_list = sorted(attributes_list, key=lambda x: x[2])

    children = []
    for i, (name, value, is_default) in enumerate(attributes_list):
        child_path = None if path is None else f"{path}.{name}"

        if i < len(attributes) - 1:
            # Not the last child. Always show a comma, and add a space when
            # collapsed.
            comma_after = rendering_parts.siblings(
                ",",
                rendering_parts.fold_condition(collapsed=rendering_parts.text(" ")),
            )
        else:
            # Last child: only show the comma when the node is expanded.
            comma_after = rendering_parts.fold_condition(
                expanded=rendering_parts.text(",")
            )

        child_line = rendering_parts.build_full_line_with_annotations(
            rendering_parts.siblings_with_annotations(
                f"{name}=",
                subtree_renderer(value, path=child_path),
            ),
            comma_after,
        )
        if is_default:
            child_line = rendering_parts.fold_condition(
                collapsed=rendering_parts.comment_color(child_line),
                expanded=rendering_parts.comment_color_when_expanded(child_line),
            )
        children.append(child_line)

    return rendering_parts.build_foldable_tree_node_from_children(
        prefix=constructor,
        children=children,
        suffix=closing_suffix,
        path=path,
        background_color=color,
    )



from __future__ import annotations

import logging
import threading
import warnings
from typing import Any, Generic, overload

from typing_extensions import TypeVar, override

log = logging.getLogger(__name__)

T = TypeVar("T", infer_variance=True)


class Singleton(Generic[T]):
    """Manages singleton access to a configuration class using the descriptor pattern.

    This class provides a way to access a singleton instance of a configuration
    class through a descriptor that automatically binds to the owner class.

    Example:
        ```python
        class MyConfig(Config):
            singleton: ClassVar[Singleton[Self]] = Singleton()
            attribute: str
        ```

    Inheritance behavior:
    - A derived class must define its own Singleton instance
    - To share a base class's singleton, explicitly assign it:
      `singleton = BaseClass.singleton`
    """

    def __init__(self) -> None:
        """Initialize the Singleton."""
        self._instance: T | None = None
        self._lock = threading.RLock()
        self._owner_cls: type[T] | None = None
        self._reference_classes: set[type[T]] = set()

    def __set_name__(self, owner: type[T], name: str) -> None:
        """Automatically bind to the owner class when assigned as a class variable."""
        if self._owner_cls is None:
            self._owner_cls = owner
        else:
            self._reference_classes.add(owner)

    def __get__(self, instance: Any, owner: type[T]) -> Singleton[T]:
        """Return the singleton instance for the owner class."""
        self._check_class_access(owner)
        return self

    def instance(self) -> T:
        """Get the singleton instance.

        Returns:
            The singleton instance of the configuration class.

        Raises:
            RuntimeError: If no instance exists yet.
            TypeError: If accessed from a derived class that doesn't explicitly assign this singleton.
        """
        self._check_class_access(None)

        if (instance := self._instance) is None:
            raise RuntimeError(
                "No instance has been initialized. Call initialize() first."
            )
        return instance

    def try_instance(self) -> T | None:
        """Try to get the singleton instance if it exists.

        Returns:
            The singleton instance of the configuration class or None if not initialized.

        Raises:
            TypeError: If accessed from a derived class that doesn't explicitly assign this singleton.
        """
        self._check_class_access(None)

        return self._instance

    @overload
    def initialize(self, instance: T, /) -> T: ...

    @overload
    def initialize(self, **kwargs: Any) -> T: ...

    def initialize(self, *args: Any, **kwargs: Any) -> T:
        """Initialize the singleton instance.

        Can either directly set an existing instance or create a new instance
        with the provided kwargs.

        Args:
            instance: An existing instance to use as the singleton (positional only).
            **kwargs: Keyword arguments to initialize a new instance (used only if
                      no positional argument is provided).

        Returns:
            The singleton instance of the configuration class.

        Raises:
            RuntimeWarning: If an instance already exists.
            TypeError: If neither instance nor kwargs are provided, or if accessed
                      from a derived class that doesn't explicitly assign this singleton.
        """
        self._check_class_access(None)

        # Quick check without acquiring the lock
        if self._instance is not None:
            warnings.warn(
                "Singleton instance already exists. "
                "The existing instance will be returned.",
                RuntimeWarning,
                stacklevel=2,
            )
            return self._instance

        # Ensure we have the managed class
        if self._owner_cls is None:
            raise TypeError(
                "Singleton is not correctly bound to a class. "
                "Ensure it is assigned as a ClassVar to the configuration class."
            )

        # If not found, acquire the lock for thread-safe initialization
        with self._lock:
            # Check again in case another thread created the instance
            if (instance := self._instance) is None:
                if len(args) == 1:
                    # Use the provided instance directly
                    instance = args[0]
                elif kwargs:
                    # Create a new instance with the provided kwargs
                    instance = self._owner_cls(**kwargs)
                else:
                    raise TypeError(
                        "Either an instance or initialization parameters must be provided."
                    )
                self._instance = instance

            return instance

    def reset(self) -> None:
        """Reset the singleton instance.

        This allows re-initialization with different parameters.

        Raises:
            TypeError: If accessed from a derived class that doesn't explicitly assign this singleton.
        """
        self._check_class_access(None)

        with self._lock:
            self._instance = None

    def _check_class_access(self, cls: type[T] | None) -> None:
        """Check if the class accessing this singleton is the owner or has explicitly assigned it."""
        if self._owner_cls is None:
            raise TypeError(
                "Singleton is not correctly bound to a class. "
                "Ensure it is assigned as a ClassVar to the configuration class."
            )

        if cls is not None and cls is not self._owner_cls:
            # Check for the behavior like the following:
            # class DerivedConfig(BaseConfig):
            #     singleton = BaseConfig.singleton
            # This allows derived classes to access the base class's singleton
            # without raising an error.
            if cls in self._reference_classes:
                log.info(
                    f"Class {cls.__name__} is accessing a singleton from {self._owner_cls.__name__}."
                )
            # Otherwise, raise an error if the class is not the owner or a reference
            # class.
            else:
                raise TypeError(
                    f"Class {cls.__name__} is not allowed to access the singleton of {self._owner_cls.__name__}."
                    " Ensure it is explicitly assigned in the derived class."
                )

    @override
    def __repr__(self) -> str:
        """Representation of the Singleton descriptor."""
        owner_name = self._owner_cls.__name__ if self._owner_cls else None
        status = "initialized" if self._instance is not None else "uninitialized"
        return f"<{self.__class__.__name__} owner={owner_name}, status={status}, instance={self._instance}>"



from __future__ import annotations

import logging
import typing
from dataclasses import dataclass
from typing import TYPE_CHECKING, Annotated, Any, ClassVar, Literal, cast

from pydantic import BaseModel, ConfigDict, Field, GetCoreSchemaHandler
from pydantic_core import CoreSchema, PydanticCustomError
from typing_extensions import TypeVar

log = logging.getLogger(__name__)


class MissingValue(BaseModel):
    model_config: ClassVar[ConfigDict] = {
        "strict": True,
        "frozen": True,
        "extra": "forbid",
    }

    NSHCONFIG___MISSING_SENTINEL: Literal["NSHCONFIG___MISSING_SENTINEL_VALUE"] = Field(
        default="NSHCONFIG___MISSING_SENTINEL_VALUE",
        title="Missing",
        # Different Pydantic versions serialize the JSON schema of this kind of
        # field differently. All versions emit {"const": "NSHCONFIG___MISSING_SENTINEL_VALUE"},
        # but some versions also emit {"type": "string"}, and some also emit
        # {"enum": ["NSHCONFIG___MISSING_SENTINEL_VALUE"]}.
        # We want to ensure that the JSON schema is always the same, so we
        # explicitly set the type and enum here, even for versions that don't
        # normally emit them.
        json_schema_extra={
            "type": "string",
            "enum": ["NSHCONFIG___MISSING_SENTINEL_VALUE"],
        },
    )


MISSING = cast(Any, MissingValue())
"""A sentinel value to indicate a missing field.
This is used to indicate that a field is missing from a model.
This is a valid value for any field that has the `AllowMissing` annotation."""


def _singleton_missing_config_new(cls, *args, **kwargs):
    """Always returns the existing MISSING singleton instance."""
    global MISSING
    log.debug(f"Intercepted {cls.__name__}() call, returning singleton.")
    return MISSING


if not TYPE_CHECKING:
    MissingValue.__new__ = classmethod(_singleton_missing_config_new)

T = TypeVar("T", infer_variance=True)

if TYPE_CHECKING:
    # If we add configurable attributes to AllowMissing, we'd probably need to stop hiding it from type checkers like this
    AllowMissing = Annotated[T, ...]
    # `AllowMissing[Sequence]` will be recognized by type checkers as `Sequence`

    _AllowMissing = object
else:

    @dataclass(frozen=True)
    class _AllowMissing:
        def __get_pydantic_core_schema__(
            self,
            source_type: Any,
            handler: GetCoreSchemaHandler,
        ) -> CoreSchema:
            return handler.generate_schema(typing.Union[source_type, MissingValue])

        def __getitem__(self, item: T) -> T:
            return cast(T, Annotated[item, self])

        def __call__(self):
            return self

    AllowMissing = _AllowMissing()


def validate_no_missing(model: BaseModel):
    for name, field in type(model).model_fields.items():
        # If the field doesn't have the `AllowMissing` annotation, ignore it.
        #   (i.e., just let Pydantic do its thing).
        if not any(isinstance(m, _AllowMissing) for m in field.metadata):
            log.debug(
                f"Skipping field '{name}' as it does not have AllowMissing annotation."
            )
            continue

        log.debug(f"Validating field '{name}' for missing value.")

        # By this point, the field **should** have some value.
        if not hasattr(model, name):
            raise PydanticCustomError(
                "field_not_set",
                'Field "{name}" is missing from the model.',
                {"name": name},
            )

        # Now, we error out if the field is missing.
        if getattr(model, name) is MISSING:
            raise PydanticCustomError(
                "field_MISSING",
                'Field "{name}" is still `MISSING`. Please provide a value for it.',
                {"name": name},
            )



from __future__ import annotations

from ._src.adapter import Adapter as Adapter
from ._src.config import Config as Config
from ._src.config import ConfigDict as ConfigDict
from ._src.config import with_config as with_config
from ._src.export import Export as Export
from ._src.invalid import Invalid as Invalid
from ._src.missing import MISSING as MISSING
from ._src.missing import AllowMissing as AllowMissing
from ._src.missing import validate_no_missing as validate_no_missing
from ._src.pydantic_exports import *
from ._src.registry import Registry as Registry
from ._src.registry import RegistryConfig as RegistryConfig
from ._src.singleton import Singleton as Singleton
from ._src.utils import deduplicate as deduplicate
from ._src.utils import deduplicate_configs as deduplicate_configs

try:
    from importlib.metadata import PackageNotFoundError, version
except ImportError:
    # For Python <3.8
    from importlib_metadata import (  # pyright: ignore[reportMissingImports]
        PackageNotFoundError,
        version,
    )

try:
    __version__ = version(__name__)
except PackageNotFoundError:
    __version__ = "unknown"



from __future__ import annotations

import contextlib
import copy
import importlib.util
import json
import logging
from collections.abc import Callable, Mapping, MutableMapping
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar, Literal, cast, get_origin, overload

from pydantic import (
    BaseModel,
    PrivateAttr,
    ValidationInfo,
    model_serializer,
    model_validator,
)
from pydantic import ConfigDict as _ConfigDict
from typing_extensions import Self, TypedDict, TypeVar, Unpack, override

from .utils import PYDANTIC_VERSION, IncEx

try:
    from pydantic import with_config as _pydantic_with_config  # type: ignore
except ImportError:
    _WithConfigTypeT = TypeVar("_WithConfigTypeT", bound=type)

    def _pydantic_with_config(
        config: ConfigDict,
    ) -> Callable[[_WithConfigTypeT], _WithConfigTypeT]:
        def inner(TypedDictClass: _WithConfigTypeT, /) -> _WithConfigTypeT:
            TypedDictClass.__pydantic_config__ = config
            return TypedDictClass

        return inner


if TYPE_CHECKING:
    from ruamel.yaml import YAML


log = logging.getLogger(__name__)

_MutableMappingBase = MutableMapping[str, Any]
if TYPE_CHECKING:
    _MutableMappingBase = object


class ConfigDict(_ConfigDict, total=False):
    repr_diff_only: bool
    """
    If `True`, the repr methods will only show values for fields that are different from the default.
    Defaults to `False`.
    """

    no_validate_assignment_for_draft: bool
    """
    Whether to disable the validation of assignments for draft configs.
    Defaults to `True`.
    """

    set_default_hash: bool
    """
    Whether to set the default hash function for the config class. By default,
    Pydantic adds a `__hash__` method for frozen models, but Config classes are mutable,
    so we don't get a hash function by default. Setting this to `True` will enable the hash function
    for the config class, which is useful for using the config class as a key in a dictionary.
    Defaults to `True`.
    """

    disable_typed_dict_generation: bool
    """
    Whether to disable the generation of TypedDict classes for the config class.
    If `True`, the `nshconfig-export` will not generate TypedDict classes for the config class.
    Defaults to `False`.
    """


class DumpKwargs(TypedDict, total=False):
    include: IncEx | None
    """Field(s) to include in the JSON output."""
    exclude: IncEx | None
    """Field(s) to exclude from the JSON output."""
    context: Any | None
    """Additional context to pass to the serializer."""
    by_alias: bool
    """Whether to serialize using field aliases."""
    exclude_unset: bool
    """Whether to exclude fields that have not been explicitly set."""
    exclude_defaults: bool
    """Whether to exclude fields that are set to their default value."""
    exclude_none: bool
    """Whether to exclude fields that have a value of `None`."""
    round_trip: bool
    """If True, dumped values should be valid as input for non-idempotent types such as Json[T]."""
    warnings: bool | Literal["none", "warn", "error"]
    """How to handle serialization errors. False/"none" ignores them, True/"warn" logs errors, "error" raises a [`PydanticSerializationError`][pydantic_core.PydanticSerializationError]."""
    serialize_as_any: bool
    """Whether to serialize fields with duck-typing serialization behavior."""


class YamlDumpKwargs(DumpKwargs, total=False):
    default_flow_style: bool | None
    """Whether to use "flow style" (more human-readable).
    https://yaml.readthedocs.io/en/latest/detail.html?highlight=default_flow_style#indentation-of-block-sequences
    """
    map_indent: int | None
    """Indent value for mappings."""
    sequence_indent: int | None
    """Indent value for sequences."""
    sequence_dash_offset: int | None
    """Indent value for the dash in sequences."""
    custom_yaml_writer: "YAML | None"
    """An instance of ruamel.yaml.YAML (or a subclass) to use as the writer.
    The above options will be set on it, if given."""


_DEFAULT_JSON_DUMP_KWARGS: DumpKwargs = {}
_DEFAULT_YAML_DUMP_KWARGS: YamlDumpKwargs = {
    "sequence_dash_offset": 0,
    # For some reason, ^ is necessary to get a properly parsable YAML output
}


_JsonLoadContextSentinel = object()

_model_config: ConfigDict = {
    # Pydantic's default config options
    **BaseModel.model_config,
    # Our overrides
    "validate_assignment": True,
    "validate_return": True,
    "validate_default": True,
    "strict": True,
    "revalidate_instances": "always",
    "arbitrary_types_allowed": True,
    "extra": "forbid",
    # Our custom config options
    "repr_diff_only": False,
    "no_validate_assignment_for_draft": True,
    "set_default_hash": True,
    "disable_typed_dict_generation": False,
}


class Config(BaseModel, _MutableMappingBase):
    """
    A base configuration class that provides validation and serialization capabilities.

    This class extends Pydantic's BaseModel and implements a mutable mapping interface. It supports draft configurations
    for flexible initialization and provides various serialization methods.
    """

    _is_draft_config: bool = PrivateAttr(default=False)
    """
    Whether this config is a draft config or not.

    Draft configs are configs that are not yet fully validated.
    They allow for a nicer API when creating configs, e.g.:

        ```python
        config = MyConfig.draft()

        # Set some values
        config.a = 10
        config.b = "hello"

        # Finalize the config
        config = config.finalize()
        ```
    """

    model_config: ClassVar[ConfigDict] = {  # pyright: ignore[reportIncompatibleVariableOverride]
        # Pydantic's default config options
        **BaseModel.model_config,
        # Our overrides
        "validate_assignment": True,
        "validate_return": True,
        "validate_default": True,
        "strict": True,
        "revalidate_instances": "always",
        "arbitrary_types_allowed": True,
        "extra": "forbid",
        # The settings below are > 2.0.0; to prevent type-checking errors on
        # older versions, we use `cast` to ensure compatibility.
        **cast(
            ConfigDict,
            {
                "validation_error_cause": True,
                "use_attribute_docstrings": True,
            },
        ),
        # Our custom config options
        "repr_diff_only": False,
        "no_validate_assignment_for_draft": True,
        "set_default_hash": True,
        "disable_typed_dict_generation": False,
    }

    if TYPE_CHECKING:

        @override
        def __init_subclass__(cls, **kwargs: Unpack[ConfigDict]) -> None:
            super().__init_subclass__(**kwargs)

    @override
    @classmethod
    def __pydantic_init_subclass__(cls, **kwargs) -> None:
        super().__pydantic_init_subclass__(**kwargs)

        if cls.model_config.get("set_default_hash", True):
            _try_set_hash(cls)

        from .registry import Registry, extract_registries_from_field_info

        log.debug(f"__pydantic_init_subclass__: {cls}")

        # Look through fields for registry annotations including nested types
        registries: list[Registry] = []
        registry_ids: set[int] = set()
        for name, field in cls.model_fields.items():
            if not (found_registries := extract_registries_from_field_info(field)):
                continue

            for registry in found_registries:
                if (id_ := id(registry)) in registry_ids:
                    continue

                log.debug(f"__pydantic_init_subclass__@{cls} - {name}: {registry}")
                registries.append(registry)
                registry_ids.add(id_)

        # Register rebuild callback with each registry found
        for registry in registries:
            registry._rebuild_on_registers_if_auto_rebuild(cls)

    def __draft_pre_init__(self):
        """Called right before a draft config is finalized."""
        pass

    def __post_init__(self):
        """Called after the final config is validated."""
        pass

    def __after_post_init__(self):
        """Called after __post_init__ is successfully called."""
        pass

    def model_deep_validate(self, strict: bool = True):
        """
        Validate the config and all of its sub-configs.

        Args:
            config: The config to validate.
            strict: Whether to validate the config strictly.
        """
        config_dict = self.model_dump(round_trip=True)
        config = self.model_validate(config_dict, strict=strict)

        return config

    @classmethod
    def draft(cls, **kwargs):
        config = cls.model_construct_draft(**kwargs)
        return config

    def finalize(self, strict: bool = True):
        # This must be a draft config, otherwise we raise an error
        if not self._is_draft_config:
            raise ValueError("Finalize can only be called on drafts.")

        # First, we call `__draft_pre_init__` to allow the config to modify itself a final time
        self.__draft_pre_init__()

        # Then, we dump the config to a dict and then re-validate it
        return self.model_deep_validate(strict=strict)

    _patched_post_init: ClassVar[bool] = False

    @classmethod
    @contextlib.contextmanager
    def _nshconfig_patch_model_post_init(cls):
        cls._patched_post_init = True
        try:
            yield
        finally:
            cls._patched_post_init = False

    @override
    def model_post_init(self, __context: Any) -> None:
        super().model_post_init(__context)

        # Call the `__post_init__` method if this is not a draft config
        if self._patched_post_init:
            return

        self.__post_init__()

        self.__after_post_init__()

    def model_validate_no_missing(self):
        from .missing import validate_no_missing

        validate_no_missing(self)

    @classmethod
    def model_construct_draft(cls, _fields_set: set[str] | None = None, **values: Any):
        """
        This is a copy of the `model_construct` method from Pydantic's `Model` class,
            with the following changes:
            - The `model_post_init` method is called with the `_DraftConfigContext` context.
            - The `_is_draft_config` attribute is set to `True` in the `values` dict.
        """

        values = copy.deepcopy(values)
        values["_is_draft_config"] = True
        with cls._nshconfig_patch_model_post_init():
            return cls.model_construct(_fields_set, **values)

    @contextlib.contextmanager
    def _nshconfig_patch_validator_validate_assignment(self):
        prev_value = self.model_config.get("validate_assignment", _notset := object())
        try:
            # We temporarily disable the validation of assignments
            self.model_config["validate_assignment"] = False
            yield
        finally:
            # We re-enable the validation of assignments
            if prev_value is _notset:
                del self.model_config["validate_assignment"]
            else:
                self.model_config["validate_assignment"] = cast(Any, prev_value)

    if not TYPE_CHECKING:

        @override
        def __setattr__(self, name: str, value: Any) -> None:
            __tracebackhide__ = True

            with contextlib.ExitStack() as stack:
                if self._is_draft_config and self.model_config.get(
                    "no_validate_assignment_for_draft", True
                ):
                    stack.enter_context(
                        self._nshconfig_patch_validator_validate_assignment()
                    )
                return super().__setattr__(name, value)

    @override
    def __repr_args__(self):
        # If `repr_diff_only` is `True`, we only show the fields that are different from the default.
        if not self.model_config.get("repr_diff_only", False):
            yield from super().__repr_args__()
            return

        # First, we get the default values for all fields.
        default_values = self.model_construct_draft()

        # Then, we compare the default values with the current values.
        for k, v in super().__repr_args__():
            if k is None:
                yield k, v
                continue

            # If there is no default value or the value is different from the default, we yield it.
            if not hasattr(default_values, k) or getattr(default_values, k) != v:
                yield k, v
                continue

            # Otherwise, we can skip this field.

    # region MutableMapping implementation
    if not TYPE_CHECKING:
        # This is mainly so the config can be used with lightning's hparams
        #   transparently and without any issues.

        @property
        def _nshconfig_dict(self):
            return self.model_dump()

        # We need to make sure every config class
        #   is a MutableMapping[str, Any] so that it can be used
        #   with lightning's hparams.
        @override
        def __getitem__(self, key: str):
            # Key can be of the format "a.b.c"
            #   so we need to split it into a list of keys.
            [first_key, *rest_keys] = key.split(".")
            value = self._nshconfig_dict[first_key]

            for key in rest_keys:
                if isinstance(value, Mapping):
                    value = value[key]
                else:
                    value = getattr(value, key)

            return value

        @override
        def __setitem__(self, key: str, value: Any):
            # Key can be of the format "a.b.c"
            #   so we need to split it into a list of keys.
            [first_key, *rest_keys] = key.split(".")
            if len(rest_keys) == 0:
                self._nshconfig_dict[first_key] = value
                return

            # We need to traverse the keys until we reach the last key
            #   and then set the value
            current_value = self._nshconfig_dict[first_key]
            for key in rest_keys[:-1]:
                if isinstance(current_value, Mapping):
                    current_value = current_value[key]
                else:
                    current_value = getattr(current_value, key)

            # Set the value
            if isinstance(current_value, MutableMapping):
                current_value[rest_keys[-1]] = value
            else:
                setattr(current_value, rest_keys[-1], value)

        @override
        def __delitem__(self, key: str):
            # This is unsupported for this class
            raise NotImplementedError

        @override
        def __iter__(self):
            return iter(self._nshconfig_dict)

        @override
        def __len__(self):
            return len(self._nshconfig_dict)

    # endregion

    # region `treescope` integration
    def __treescope_repr__(self, path, subtree_renderer):
        if importlib.util.find_spec("treescope") is None:
            raise ImportError("The `treescope` package is required for this feature.")

        from .treescope_util import render_object_constructor

        # For the attributes, let's first output the fields that are different from the default.
        attributes = {}

        # First, we get the default values for all fields.
        default_values = self.model_construct_draft()

        # Then, we compare the default values with the current values.
        for k, v in super().__repr_args__():
            if k is None:
                continue

            is_default = hasattr(default_values, k) and getattr(default_values, k) == v
            attributes[k] = (v, is_default)

        return render_object_constructor(
            object_type=type(self),
            attributes=attributes,
            path=path,
            subtree_renderer=subtree_renderer,
            roundtrippable=True,
        )

    # endregion
    @model_serializer(mode="wrap")
    def include_literals(self, next_serializer):
        """Include fields with the `Literal` annotation in the dumped data,
        even if `exclude_defaults` is `True`.

        See https://github.com/pydantic/pydantic/discussions/9108#discussioncomment-8926452
        for the original implementation."""
        dumped = next_serializer(self)
        for name, field_info in type(self).model_fields.items():
            if get_origin(field_info.annotation) == Literal:
                dumped[name] = getattr(self, name)
        return dumped

    # region Construction methods
    # Dict
    @classmethod
    def from_dict(cls, config_dict: dict[str, Any]):
        """Create configuration from a dictionary.

        Args:
            config_dict: Dictionary containing configuration values

        Returns:
            A validated configuration instance
        """
        return cls.model_validate(config_dict)

    @classmethod
    def from_dict_or_instance(cls, data: Mapping[str, Any] | Self):
        """Create configuration from a dictionary or an instance.

        Args:
            data: Dictionary containing configuration values or an instance of the config

        Returns:
            A validated configuration instance
        """
        if isinstance(data, cls):
            return data
        elif isinstance(data, Mapping):
            return cls.from_dict(dict(data))
        else:
            raise TypeError(
                "data must be a dictionary or an instance of the config class"
            )

    def to_dict(self) -> dict[str, Any]:
        """Convert configuration to a dictionary.

        Returns:
            Dictionary representation of the configuration
        """
        return self.model_dump()

    # JSON

    @model_validator(mode="before")
    @classmethod
    def _pop_schema(cls, values: dict[str, Any], info: ValidationInfo):
        if (
            not info.context
            or not isinstance(info.context, dict)
            or (json_load_context := info.context.get("_nshconfig_json_load")) is None
            or json_load_context is not _JsonLoadContextSentinel
        ):
            return values

        # If the loaded config contains a `$schema` key (which was added by the
        # `to_json_str` method with `with_schema=True`), we remove it.
        values.pop("$schema", None)
        return values

    def to_json_str(
        self,
        /,
        with_schema: bool = True,
        indent: int | None = 4,
        **kwargs: Unpack[DumpKwargs],
    ) -> str:
        """Dump configuration to a JSON string.

        Args:
            with_schema: Whether to include the schema reference in the JSON file
            indent: Number of spaces to use for indentation
            kwargs: Additional keyword arguments to pass to the YAML dumper,
                these are parameters directly passed to cls.model_dump_json().
        """
        kwargs_with_defaults = _DEFAULT_JSON_DUMP_KWARGS.copy()
        kwargs_with_defaults.update(kwargs)
        json_str = self.model_dump_json(
            indent=indent, **cast(Any, kwargs_with_defaults)
        )

        # Add schema reference if available and with_schema is True
        if with_schema and (schema_uri := _get_schema_uri(type(self))):
            try:
                # A bit hacky here, but we will re-load the JSON,
                # add the $schema key, and then dump it again.
                json_obj = json.loads(json_str)
                if isinstance(json_obj, MutableMapping):
                    json_obj["$schema"] = schema_uri
                    json_str = json.dumps(json_obj, indent=indent)
                else:
                    log.warning(
                        "Could not add schema reference to JSON string. "
                        "The JSON object is not a dictionary. "
                        "The schema reference will not be added."
                    )
            except json.JSONDecodeError:
                log.warning(
                    "Could not add schema reference to JSON string. "
                    "The JSON string is not valid JSON. "
                    "The schema reference will not be added."
                )

        return json_str

    def to_json_file(
        self,
        path: str | Path,
        /,
        with_schema: bool = True,
        indent: int | None = 4,
        **kwargs: Unpack[DumpKwargs],
    ) -> None:
        """Save configuration to a JSON file.

        Args:
            path: Path where the JSON file will be saved
            with_schema: Whether to include the schema reference in the JSON file
            indent: Number of spaces to use for indentation
            kwargs: Additional keyword arguments to pass to the YAML dumper,
                these are parameters directly passed to cls.model_dump_json().
        """
        json_str = self.to_json_str(with_schema=with_schema, indent=indent, **kwargs)
        with open(path, "w", encoding="utf-8") as f:
            f.write(json_str)

    @classmethod
    def from_json_str(cls, json_str: str | bytes, /):
        """Create configuration from a JSON string.

        Args:
            json_str: JSON string

        Returns:
            A validated configuration instance
        """
        return cls.model_validate_json(
            json_str, context={"_nshconfig_json_load": _JsonLoadContextSentinel}
        )

    @classmethod
    def from_json_file(cls, path: str | Path, /):
        """Create configuration from a JSON file.

        Args:
            path: Path to the JSON file

        Returns:
            A validated configuration instance
        """
        with open(path, "r", encoding="utf-8") as f:
            return cls.from_json_str(f.read())

    def to_yaml_str(
        self,
        /,
        with_schema: bool = True,
        indent: int | None = 4,
        **kwargs: Unpack[YamlDumpKwargs],
    ) -> str:
        """Dump configuration to a YAML string.

        Args:
            with_schema: Whether to include the schema reference in the YAML string
            indent: Number of spaces to use for indentation
            kwargs: Additional keyword arguments to pass to the YAML dumper,
                these are parameters directly passed to cls.model_dump_json().

        Raises:
            ImportError: If pydantic-yaml is not installed
        """
        try:
            from pydantic_yaml import to_yaml_str
        except ImportError:
            raise ImportError(
                "Pydantic-yaml is required for YAML support. "
                "You can either install nshconfig with "
                "all extras using 'pip install nshconfig[extra]'"
                ", install the yaml extra via 'pip install nshconfig[yaml]'"
                ", or install with 'pip install pydantic-yaml'"
            )

        kwargs_with_defaults = _DEFAULT_YAML_DUMP_KWARGS.copy()
        kwargs_with_defaults.update(kwargs)
        data_str = to_yaml_str(self, indent=indent, **kwargs_with_defaults)

        # Add YAML language server schema directive if with_schema is True
        if with_schema and (schema_uri := _get_schema_uri(type(self))):
            data_str = f"# yaml-language-server: $schema={schema_uri}\n\n" + data_str

        return data_str

    def to_yaml_file(
        self,
        path: str | Path,
        /,
        with_schema: bool = True,
        indent: int | None = 4,
        **kwargs: Unpack[YamlDumpKwargs],
    ) -> None:
        """Save configuration to a YAML file.

        Args:
            path: Path where the YAML file will be saved
            with_schema: Whether to include the schema reference in the YAML file
            indent: Number of spaces to use for indentation
            kwargs: Additional keyword arguments to pass to the YAML dumper,
                these are parameters directly passed to cls.model_dump_json().

        Raises:
            ImportError: If pydantic-yaml is not installed
        """
        yaml_str = self.to_yaml_str(with_schema=with_schema, indent=indent, **kwargs)

        # Write the file
        with open(path, "w", encoding="utf-8") as f:
            f.write(yaml_str)

    @classmethod
    def from_yaml_str(cls, yaml_str: str, /):
        """Create configuration from a YAML string.

        Args:
            yaml_str: YAML string

        Returns:
            A validated configuration instance

        Raises:
            ImportError: If pydantic-yaml is not installed
        """
        try:
            from pydantic_yaml import parse_yaml_raw_as
        except ImportError:
            raise ImportError(
                "Pydantic-yaml is required for YAML support. "
                "You can either install nshconfig with "
                "all extras using 'pip install nshconfig[extra]'"
                ", install the yaml extra via 'pip install nshconfig[yaml]'"
                ", or install with 'pip install pydantic-yaml'"
            )

        return parse_yaml_raw_as(cls, yaml_str)

    @classmethod
    def from_yaml(cls, path: str | Path, /):
        """Create configuration from a YAML file.

        Args:
            path: Path to the YAML file

        Returns:
            A validated configuration instance

        Raises:
            ImportError: If pydantic-yaml is not installed
        """
        try:
            from pydantic_yaml import parse_yaml_file_as
        except ImportError:
            raise ImportError(
                "Pydantic-yaml is required for YAML support. "
                "You can either install nshconfig with "
                "all extras using 'pip install nshconfig[extra]'"
                ", install the yaml extra via 'pip install nshconfig[yaml]'"
                ", or install with 'pip install pydantic-yaml'"
            )

        return parse_yaml_file_as(cls, path)

    @classmethod
    def from_python_file(cls, path: str | Path, /):
        """Create configuration from a Python file.

        The Python file should export a `__config__` variable that contains the configuration,
        or a `__create_config__` function that returns a configuration.

        Args:
            path: Path to the Python file

        Returns:
            A validated configuration instance

        Raises:
            FileNotFoundError: If the Python file does not exist
            ImportError: If the Python file cannot be imported
            ValueError: If the Python file does not export a valid `__config__` variable
        """
        from .utils import import_and_parse_python_file

        return import_and_parse_python_file(path, cls)

    @classmethod
    def from_python_module(cls, module_name: str, /):
        """Create configuration from a Python module.

        The Python module should export a `__config__` variable that contains the configuration,
        or a `__create_config__` function that returns a configuration.

        Args:
            module_name: Name of the Python module (e.g. "myapp.config")

        Returns:
            A validated configuration instance

        Raises:
            ImportError: If the Python module cannot be imported
            ValueError: If the Python module does not export a valid `__config__` variable
        """
        from .utils import import_and_parse_python_module

        return import_and_parse_python_module(module_name, cls)

    # endregion

    def to_toml_str(
        self,
        /,
        indent: int = 4,
        multiline_strings: bool = False,
        **kwargs: Unpack[DumpKwargs],
    ) -> str:
        """Dump configuration to a TOML string.

        Args:
            indent: Number of spaces to use for indentation
            multiline_strings: Whether to use multiline strings
            kwargs: Additional keyword arguments to pass to the TOML dumper,
                these are parameters directly passed to cls.model_dump_json().

        Raises:
            ImportError: If pydantic-yaml is not installed
        """

        try:
            from tomli_w import dumps
        except ImportError:
            raise ImportError(
                "Tomli-w is required for TOML support. "
                "You can either install nshconfig with "
                "all extras using 'pip install nshconfig[extra]'"
                ", install the w extra via 'pip install nshconfig[toml]'"
                ", or install with 'pip install tomli-w'"
            )

        # We need to convert the config to a dict first
        config_dict = self.model_dump(**cast(Any, kwargs))

        # Then we can use tomli_w to dump the dict to a TOML string
        toml_str = dumps(
            config_dict, indent=indent, multiline_strings=multiline_strings
        )
        return toml_str

    def to_toml_file(
        self,
        /,
        path: str | Path,
        indent: int = 4,
        multiline_strings: bool = False,
        **kwargs: Unpack[DumpKwargs],
    ) -> None:
        """Dump configuration to a TOML file.

        Args:
            path: Path to the TOML file
            indent: Number of spaces to use for indentation
            multiline_strings: Whether to use multiline strings
            kwargs: Additional keyword arguments to pass to the TOML dumper,
                these are parameters directly passed to cls.model_dump_json().
        """
        toml_str = self.to_toml_str(
            indent=indent, multiline_strings=multiline_strings, **kwargs
        )

        with open(path, "w") as f:
            f.write(toml_str)

    @classmethod
    def from_toml_str(cls, toml_str: str, /):
        """Create configuration from a TOML string.

        Args:
            toml_str: TOML string

        Returns:
            A validated configuration instance

        Raises:
            ImportError: If tomli is not installed
        """
        try:
            from tomli import loads
        except ImportError:
            raise ImportError(
                "Tomli is required for TOML support. "
                "You can either install nshconfig with "
                "all extras using 'pip install nshconfig[extra]'"
                ", install the tomli extra via 'pip install nshconfig[toml]'"
                ", or install with 'pip install tomli'"
            )

        config_dict = loads(toml_str)
        return cls.from_dict(config_dict)

    @classmethod
    def from_toml_file(cls, path: str | Path, /):
        """Create configuration from a TOML file.

        Args:
            path: Path to the TOML file
        """
        toml_str = Path(path).read_text()
        return cls.from_toml_str(toml_str)


def _get_schema_uri(cls: type[Config]) -> str | None:
    """Helper to get the absolute schema path for a config class."""
    # First check for __nshconfig_json_schema_uri__ directly on the class (not inherited)
    if "__nshconfig_json_schema_uri__" in cls.__dict__:
        return getattr(cls, "__nshconfig_json_schema_uri__")

    from .export import find_config_metadata

    if not (metadata := find_config_metadata(cls)):
        return None

    _, schema_path = metadata
    if schema_path is None:
        return None

    # Get the absolute path by looking up from the metadata file location
    if (
        spec := importlib.util.find_spec(cls.__module__)
    ) is None or spec.origin is None:
        return None

    module_dir = Path(spec.origin).parent
    # Look up until we find the .nshconfig.generated.json file
    current_dir = module_dir
    while current_dir.parent != current_dir:
        if (current_dir / ".nshconfig.generated.json").exists():
            file_path = str((current_dir / schema_path).resolve().absolute())
            return f"file://{file_path}"
        current_dir = current_dir.parent

    return None


def _try_set_hash_legacy(cls: type[Config], bases: tuple[type, ...]):
    if "__hash__" in cls.__dict__:
        return

    from pydantic_core import PydanticUndefined

    base_hash_func = None
    for base in bases:
        base_hash_func = getattr(base, "__hash__", PydanticUndefined)
        if base_hash_func is not PydanticUndefined:
            break

    if base_hash_func is None:
        # This will be the case for `BaseModel` since it defines `__eq__` but not `__hash__`.
        # In this case, we generate a standard hash function, generally for use with frozen models.

        def hash_func(self: Any) -> int:
            return hash(self.__class__) + hash(tuple(self.__dict__.values()))

        cls.__hash__ = hash_func  # type: ignore[assignment]


def _try_set_hash(cls: type[Config]):
    # On earlier versions of Pydantic, `set_default_hash_func` has a different signature,
    # so we need to handle that gracefully.
    if PYDANTIC_VERSION < "2.6":
        return _try_set_hash_legacy(cls, cls.__bases__)

    # This is a bit of a hack as we're relying on Pydantic's internal API,
    # but we can fail gracefully if it doesn't work.

    try:
        from pydantic._internal._model_construction import (
            set_default_hash_func,  # type: ignore
        )
    except ImportError:
        log.warning(
            "Could not set default hash function for config class. "
            "This should not happen. Please report this issue."
        )
        return

    set_default_hash_func(cast(Any, cls), cls.__bases__)
    log.debug(f"Default hash function set for class: {cls.__name__}")


_TypeT = TypeVar("_TypeT", bound=type[Config], infer_variance=True)


@overload
def with_config(config: ConfigDict, /) -> Callable[[_TypeT], _TypeT]: ...


@overload
def with_config(**config: Unpack[ConfigDict]) -> Callable[[_TypeT], _TypeT]: ...


def with_config(
    config: ConfigDict | None = None, /, **kwargs: Any
) -> Callable[[_TypeT], _TypeT]:
    return _pydantic_with_config(config, **kwargs)  # pyright: ignore[reportArgumentType]



from __future__ import annotations

from typing import TYPE_CHECKING, cast

import pydantic as pydantic  # full module export
from pydantic import *  # pyright: ignore[reportWildcardImportFromLibrary]
from typing_extensions import Never

# Delete the BaseModel and create_model imports from pydantic
if TYPE_CHECKING:
    BaseModel = cast(Never, None)
    create_model = cast(Never, None)
else:
    del BaseModel
    del create_model

# Re-export ConfigDict and with_config from our implementation
if not TYPE_CHECKING:
    del ConfigDict
    try:
        del with_config
    except NameError:
        pass
from .config import ConfigDict as ConfigDict
from .config import with_config as with_config



from __future__ import annotations

import dataclasses
import logging
import typing
import warnings
from collections.abc import Callable, Iterable
from typing import TYPE_CHECKING, Any, Generic, Literal, TypedDict, TypeVar, cast

from pydantic import GetCoreSchemaHandler, model_validator
from pydantic.fields import FieldInfo
from pydantic_core import core_schema
from typing_extensions import assert_never, deprecated, override

from .utils import PYDANTIC_VERSION

if TYPE_CHECKING:
    from .config import Config, ConfigDict

log = logging.getLogger(__name__)

TConfig = TypeVar("TConfig", bound="Config", covariant=True)
TClass = TypeVar("TClass", bound="type[Config]")


@dataclasses.dataclass(frozen=True)
class _RegistryEntry:
    tag: str
    cls: type[Config]


def _unwrap_annotated_recurive(typ: Any):
    while typing.get_origin(typ) is typing.Annotated:
        typ = typing.get_args(typ)[0]
    return typ


def _resolve_tag(cls: type[Config], discriminator_field_name: str) -> str:
    # Make sure cls has the discriminator attribute
    if (discriminator_field := cls.model_fields.get(discriminator_field_name)) is None:
        raise ValueError(f"{cls} does not have a field `{discriminator_field_name}`")

    # Make sure the discriminator field is a Literal
    if (
        typing.get_origin(
            type_ := _unwrap_annotated_recurive(discriminator_field.annotation)
        )
        is not typing.Literal
    ):
        raise ValueError(
            f"The discriminator field `{discriminator_field_name}` of {cls} "
            f"should be a Literal. Got {type_}. Please ensure the field is a Literal."
        )

    # Make sure the Literal has only one argument
    if len(typing.get_args(type_)) != 1:
        raise ValueError(
            f"The discriminator field `{discriminator_field_name}` of {cls} "
            f"should be a Literal with only one argument. Got {type_}. "
            "This should not happen. Please report a bug."
        )

    # Extract the tag
    return typing.get_args(type_)[0]


class RegistryConfig(TypedDict, total=False):
    """A typed dictionary defining configuration options for the registry.

    Attributes:
        duplicate_tag_policy (Literal["warn-and-ignore", "warn-and-replace", "error"]):
            Defines how to handle duplicate tags in the registry.
            - "warn-and-ignore": Warns about duplicates but keeps the original tag
            - "warn-and-replace": Warns about duplicates and replaces with new tag
            - "error": Raises an error when duplicate tags are found

        auto_rebuild (bool):
            If True, the registry will automatically rebuild Pydantic models
            that use the registry type annotation whenever a new class
            is registered. This ensures that the model schema is always up-to-date
            with the registered classes. If False, the user must manually wrap
            parent pydantic models with the `@registry.rebuild_on_registers` decorator
            to achieve the same behavior. Default is True.
    """

    duplicate_tag_policy: Literal["warn-and-ignore", "warn-and-replace", "error"]
    auto_rebuild: bool


def _no_configs_registered_invalid_config(registry: Registry) -> type[Config]:
    if registry._invalid_cls is None:
        from .config import Config

        _model_config: ConfigDict = {}
        if PYDANTIC_VERSION >= "2.1.0":
            _model_config = {"defer_build": True}  # pyright: ignore[reportAssignmentType]

        class NoConfigsRegisteredInvalidConfig(Config):
            model_config = {**_model_config}

            @model_validator(mode="before")
            @classmethod
            def invalidate(cls, data: Any) -> Any:
                raise ValueError(
                    f"No configs have been registered with registry {registry}."
                )

        registry._invalid_cls = NoConfigsRegisteredInvalidConfig

    return registry._invalid_cls


# Pydantic < 2.1.0 has a behavior where in some situations, it expects
# types to be hashable. This is a workaround to ensure that
# the Registry class is hashable.
_legacy_pydantic_class_kwargs: Any = {}
_legacy_pydantic_field_kwargs: Any = {}
if PYDANTIC_VERSION < "2.1.0":
    _legacy_pydantic_class_kwargs = {"unsafe_hash": True}
    _legacy_pydantic_field_kwargs = {"hash": False}


def _compat_monkey_patch_type_pr_8526():
    from pydantic._internal import _core_utils

    def handle_definitions_schema(
        self,
        schema: core_schema.DefinitionsSchema,
        f: _core_utils.Walk,  # pyright: ignore[reportAttributeAccessIssue]
    ) -> core_schema.CoreSchema:
        new_definitions: list[core_schema.CoreSchema] = []
        for definition in schema["definitions"]:
            if "schema_ref" and "ref" in definition:
                # This indicates a purposely indirect reference
                # We want to keep such references around for implications related to JSON schema, etc.:
                new_definitions.append(definition)
                # However, we still need to walk the referenced definition:
                self.walk(definition, f)
                continue

            updated_definition = self.walk(definition, f)
            if "ref" in updated_definition:
                # If the updated definition schema doesn't have a 'ref', it shouldn't go in the definitions
                # This is most likely to happen due to replacing something with a definition reference, in
                # which case it should certainly not go in the definitions list
                new_definitions.append(updated_definition)
        new_inner_schema = self.walk(schema["schema"], f)

        if not new_definitions and len(schema) == 3:
            # This means we'd be returning a "trivial" definitions schema that just wrapped the inner schema
            return new_inner_schema

        new_schema = schema.copy()
        new_schema["schema"] = new_inner_schema
        new_schema["definitions"] = new_definitions
        return new_schema

    _core_utils._WalkCoreSchema.handle_definitions_schema = handle_definitions_schema  # type: ignore
    _core_utils._dispatch = _core_utils._WalkCoreSchema().walk  # type: ignore


@dataclasses.dataclass(**_legacy_pydantic_class_kwargs)
class Registry(Generic[TConfig]):
    """A registry system for creating dynamic discriminated unions with Pydantic models.

    This registry allows dynamic registration of subclasses that share a common
    discriminator field. It's particularly useful when you want to:
    1. Allow runtime registration of new subtypes
    2. Use a discriminator field to differentiate between subtypes
    3. Have Pydantic handle validation/serialization of the union type

    Args:
        base_cls: The base class that all registered classes must inherit from
        discriminator_field: The field name used to distinguish between different types

    Example:
        ```python
        class AnimalBaseConfig(Config, ABC):
            type: str

            @abstractmethod
            def make_sound(self) -> str: ...

        registry = Registry(AnimalBaseConfig, "type")

        @registry.register
        class DogConfig(AnimalBaseConfig):
            type: Literal["dog"]
            name: str

            @override
            def make_sound(self) -> str:
                return "Woof!"

        @registry.register
        class CatConfig(AnimalBaseConfig):
            type: Literal["cat"]
            name: str

            @override
            def make_sound(self) -> str:
                return "Meow!"
        ```

        @registry.rebuild_on_registers
        class ProgramConfig(Config):
            animal: Annotated[AnimalBaseConfig, registry]

        # ^ With the above code, the `ProgramConfig` class will have a field `animal`
        # that can be any of the registered classes in the registry. Our implementation
        # above implements Dog and Cat classes, and so the following code will work:

        def main(config: ProgramConfig):
            print(config.animal.make_sound())

        main(ProgramConfig(animal=DogConfig(type="dog", name="Buddy")))
        main(ProgramConfig(animal=CatConfig(type="cat", name="Whiskers")))

        # Now, let's say a user wants to add their own animal class and
        # use it in the `ProgramConfig` class. They can do so by simply
        # registering their class with the registry:

        @registry.register
        class CowConfig(AnimalBaseConfig):
            type: Literal["cow"]
            name: str

            @override
            def make_sound(self) -> str:
                return "Moo!"

        # The following code will now work (note that this would not have worked
        # if we didn't use the `@registry.register` decorator):
        main(ProgramConfig(animal=CowConfig(type="cow", name="Bessie")))
    """

    base_cls: type[TConfig]
    discriminator: str
    config: RegistryConfig = dataclasses.field(
        default_factory=lambda: {
            "duplicate_tag_policy": "error",
            "auto_rebuild": True,
        },
        **_legacy_pydantic_field_kwargs,
    )
    _elements: list[_RegistryEntry] = dataclasses.field(
        default_factory=lambda: [],
        **_legacy_pydantic_field_kwargs,
    )
    _on_register_callbacks: list[Callable[[type[Config]], None]] = dataclasses.field(
        default_factory=lambda: [],
        **_legacy_pydantic_field_kwargs,
    )
    _invalid_cls: type[Config] | None = None

    def __post_init__(self):
        # Pydantic versions between 2.4.2 and 2.5.3 have a bug in the way
        # they handle TypeAliasType (see issue: https://github.com/pydantic/pydantic/pull/8526/)
        # Here, we will just monkey-patch the code to apply the fix from the
        # Pydantic PR to ensure that TypeAliasType works correctly.
        if PYDANTIC_VERSION >= "2.4.2" and PYDANTIC_VERSION < "2.6.0":
            warnings.warn(
                "Pydantic versions 2.4.2 to 2.5.3 have a bug with TypeAliasType. "
                "This registry will monkey-patch the code to fix this issue. "
                "Please upgrade to Pydantic 2.6.0 or later when available.",
            )
            _compat_monkey_patch_type_pr_8526()

        # There is a bug in Pydantic > 2.6.0 (but not 2.6.0 itself) and <= 2.7.0
        # which makes complex, recursive types not work properly. We should
        # log a warning about this and suggest users to upgrade to 2.7.1 or later,
        # but there is nothing more we can do about it.
        if PYDANTIC_VERSION > "2.6.0" and PYDANTIC_VERSION <= "2.7.0":
            warnings.warn(
                "Pydantic versions > 2.6.0 and <= 2.7.0 have a bug with complex, recursive types. "
                "The registry will not work properly if you use complex types with it. "
                "Please upgrade to Pydantic 2.7.1 or later.",
            )

    def _ref(
        self,
        prefix: Iterable[str] | None = None,
        suffix: Iterable[str] | None = None,
    ):
        """Get the reference name for the registry.

        This is used for Pydantic schema generation and should be unique
        for each registry instance.
        """
        if prefix is None:
            prefix = ()
        if suffix is None:
            suffix = ()
        lhs = [
            *prefix,
            f"{self.base_cls.__module__}.{self.base_cls.__name__}Registry",
            *suffix,
        ]
        lhs = "_".join(lhs)
        rhs = [str(id(self))]
        rhs = "_".join(rhs)
        return f"{lhs}:{rhs}"

    def register(self, cls: TClass, /) -> TClass:
        """Register a new type with the registry.

        The class must:
        1. Inherit from the registry's base class
        2. Have a discriminator field with a single-value Literal type
        3. Not already be registered
        4. Use a unique tag value

        Args:
            cls: The class to register

        Returns:
            The registered class (allows use as decorator)

        Raises:
            ValueError: If registration requirements aren't met
        """

        # Check if the cls is a subclass of the base_cls
        if not issubclass(cls, self.base_cls):
            raise ValueError(f"{cls} should be a subclass of {self.base_cls}.")

        # Check if the cls is already registered
        if cls in set(e.cls for e in self._elements):
            raise ValueError(f"{cls} is already registered.")

        # Extract the tag from the cls
        tag = _resolve_tag(cls, self.discriminator)

        # Check if the tag is already registered
        if (
            registered_by_tag := next(
                (e.cls for e in self._elements if e.tag == tag), None
            )
        ) is not None and registered_by_tag != cls:
            if (
                policy := self.config.get("duplicate_tag_policy", "error")
            ) == "warn-and-ignore":
                log.warning(
                    f"Tag `{tag}` is already registered by {registered_by_tag}. Ignoring {cls}."
                )
                return cast(TClass, registered_by_tag)
            elif policy == "warn-and-replace":
                log.warning(
                    f"Tag `{tag}` is already registered by {registered_by_tag}. Replacing with {cls}."
                )
                self._elements = [e for e in self._elements if e.tag != tag]
            elif policy == "error":
                raise ValueError(
                    f"Tag `{tag}` is already registered by {registered_by_tag}."
                )
            else:
                assert_never(policy)

        # Add the cls to the registry
        self._elements.append(_RegistryEntry(tag=tag, cls=cls))
        log.info(f"Registered {cls} with tag '{tag}'.")

        # Call the on_register callbacks
        for callback in self._on_register_callbacks:
            callback(cls)

        return cast(TClass, cls)

    def rebuild_on_registers(self, cls: TClass, /) -> TClass:
        """Rebuild a class's schema whenever new types are registered.

        Since Pydantic caches model schemas, fields using the registry's union type
        won't automatically update when new types are registered. This decorator
        ensures the schema is rebuilt whenever the registry changes.

        For example, the code below will work for the first print (since `DogConfig`
        is registered before the `ProgramConfig` schema is built), but will fail
        for the second print (since `MooseConfig` is registered after the schema is built):

        ```python
        class AnimalBaseConfig(Config, ABC): ...

        registry = Registry(AnimalBaseConfig, "type")

        @registry.register
        class DogConfig(AnimalBaseConfig): ...

        @registry.register
        class CatConfig(AnimalBaseConfig): ...

        class ProgramConfig(Config):
            animal: Annotated[AnimalBaseConfig, registry]

        # This will work, since `DogConfig` is registered before the schema is built.
        print(ProgramConfig(animal=DogConfig(type="dog")))
        # Same with `CatConfig`
        print(ProgramConfig(animal=CatConfig(type="cat")))

        @registry.register
        class MooseConfig(AnimalBaseConfig): ...

        # This will fail, since the schema for `ProgramConfig` is already built
        # and does not include `MooseConfig`. I.e., the schema for
        # `ProgramConfig.animal` is that of `Union[DogConfig, CatConfig]`.
        print(ProgramConfig(animal=MooseConfig(type="moose")))

        # If we rebuild the `ProgramConfig` schema, then the second print will work.
        ProgramConfig.model_rebuild(force=True)

        # Now this will work
        print(ProgramConfig(animal=MooseConfig(type="moose")))
        ```

        This decorator is a convenience method that automatically rebuilds the class
        whenever a new class is registered with the registry. This way, the schema
        for the class will always be up-to-date with the registered classes. For
        example, the code below will work as expected:

        ```python
        class AnimalBaseConfig(Config, ABC): ...

        registry = Registry(AnimalBaseConfig, "type")

        @registry.register
        class DogConfig(AnimalBaseConfig): ...

        @registry.register
        class CatConfig(AnimalBaseConfig): ...

        @registry.rebuild_on_registers
        class ProgramConfig(Config):
            animal: Annotated[AnimalBaseConfig, registry]

        # This will work, since `DogConfig` is registered before the schema is built.
        print(ProgramConfig(animal=DogConfig(type="dog")))
        # Same with `CatConfig`
        print(ProgramConfig(animal=CatConfig(type="cat")))

        @registry.register
        class MooseConfig(AnimalBaseConfig): ...

        # Now, this will also work, since the schema for `ProgramConfig` is
        # automatically rebuilt whenever a new class is registered with the registry.
        print(ProgramConfig(animal=MooseConfig(type="moose")))

        # The schema for `ProgramConfig` is always up-to-date with the registered classes.
        ```

        Args:
            cls: The class to rebuild whenever a new class is registered with the registry. This
                class must be a `nshconfig.Config` subclass.

        Returns:
            The decorated class
        """

        def _rebuild(_: type[Config]):
            # For older Pydantic versions, the `model_rebuild` method
            # does not work properly, so we need to do some additional work.
            if PYDANTIC_VERSION < "2.1.0":
                if "__pydantic_core_schema__" in cls.__dict__:
                    delattr(
                        cls, "__pydantic_core_schema__"
                    )  # delete cached value to ensure full rebuild happens

            cls.model_rebuild(force=True, raise_errors=False)
            log.debug(f"Rebuilt {cls} schema due to registry changes.")

        log.debug(f"Registering {cls} for auto-rebuild on registry changes.")
        self._on_register_callbacks.append(_rebuild)
        _rebuild(cls)
        return cls

    def _rebuild_on_registers_if_auto_rebuild(
        self, cls: type[Config], /, *args: Any, **kwargs: Any
    ) -> type[Config]:
        """Rebuild the class schema if auto_rebuild is enabled.

        This is a helper method that checks the registry's config for auto_rebuild
        and rebuilds the class schema if it is enabled. It is used internally by
        the `rebuild_on_registers` decorator.
        """
        if self.config.get("auto_rebuild", True):
            log.debug(f"Auto-rebuilding {cls} schema due to registry changes.")
            return self.rebuild_on_registers(cls, *args, **kwargs)
        return cls

    def type_hint(self):
        """Create a type hint for the union of all registered types."""
        from pydantic import Field

        if not self._elements:
            # If no elements are registered, just use the Invalid type.
            t = _no_configs_registered_invalid_config(self)
        else:
            # Construct the annotated union type.
            t = typing.Union[tuple(e.cls for e in self._elements)]
            # Pydantic < 2.1.0 does not support discriminated Unions with 1 element,
            # so we need to handle that case separately.
            if PYDANTIC_VERSION < "2.1.0" and len(self._elements) == 1:
                field_info = Field()
            else:
                field_info = Field(discriminator=self.discriminator)
            field_info.annotation = t
            t = typing.Annotated[t, field_info]

        return t

    def type_adapter(self):
        """Create a TypeAdapter for validating against the registry's types.

        Returns:
            A TypeAdapter for validating/serializing registry types
        """
        from pydantic import TypeAdapter

        # Earlier versions of TypeAdapter do not have the proper overload
        # for TypeAdapter.__init__ that handles `Any` types (which is needed
        # in cases we are trying to pass an `Annotated` type), so we explicitly
        # cast this to Any to avoid type errors.
        type_ = cast(Any, self.type_hint())
        return TypeAdapter[TConfig](type_)

    def construct(self, config: Any) -> TConfig:
        """Construct a registered type instance from configuration data.

        Args:
            config: Configuration data to validate/construct from

        Returns:
            An instance of the appropriate registered type
        """

        return self.type_adapter().validate_python(config)

    def print_registered_tags(self):
        """Pretty prints all registered tags in the registry."""
        if not self._elements:
            output = "No tags registered."
        else:
            output = "Registered tags:\n"
            for entry in self._elements:
                output += f"  - {entry.tag}: {entry.cls.__name__}\n"

        print(output)

    @override
    def __repr__(self) -> str:
        """Return a string representation of the registry.

        Provides a concise overview of the registry configuration and its registered types.
        The representation includes the base class, discriminator field, and a summary
        of registered types.

        Returns
        -------
        str
            A formatted string representation of the registry
        """
        base_name = f"{self.base_cls.__module__}.{self.base_cls.__name__}"
        registered_types = len(self._elements)

        # Start with the basic info
        parts = [
            f"Registry[{base_name}](",
            f"  discriminator='{self.discriminator}',",
            f"  registered_types={registered_types},",
        ]

        # Add policy info
        policy = self.config.get("duplicate_tag_policy", "error")
        parts.append(f"  duplicate_tag_policy='{policy}',")

        auto_rebuild = self.config.get("auto_rebuild", True)
        parts.append(f"  auto_rebuild={auto_rebuild},")

        # Add registered tags if there are any
        if self._elements:
            parts.append("  tags=[")
            # Sort by tag for consistent output
            for entry in sorted(self._elements, key=lambda e: e.tag):
                class_name = entry.cls.__name__
                parts.append(f"    '{entry.tag}': {class_name},")
            parts.append("  ]")
        else:
            parts.append("  tags=[]")

        parts.append(")")

        return "\n".join(parts)

    @deprecated(
        "This method is deprecated. You can directly use the registry instance instead (e.g., Annotated[AnimalBase, registry]).",
    )
    def DynamicResolution(self):
        """Create a type annotation that enables dynamic type resolution using this registry.

        This method is a core part of using the registry system with Pydantic models. It creates
        a special type annotation that tells Pydantic to dynamically resolve field types using
        the registry's registered classes. This enables powerful dynamic behavior where fields
        can accept any type registered with the registry, even types registered after the model
        is first defined.

        Returns:
            A class that can be used with typing.Annotated to indicate a field should
            accept any registered type.

        Example:
            Basic usage with a single field:
            ```python
            @registry.rebuild_on_registers
            class Config(Config):
                # This field will accept any type registered with 'registry'
                animal: Annotated[AnimalBase, registry.DynamicResolution()]

            # These will all work:
            Config(animal=DogConfig(type="dog", name="Rover"))
            Config(animal=CatConfig(type="cat", name="Whiskers"))

            # Even types registered after Config is defined will work:
            @registry.register
            class BirdConfig(AnimalBase):
                type: Literal["bird"]
                name: str

            Config(animal=BirdConfig(type="bird", name="Tweety"))  # Works!
            ```

        Important Notes:
            1. Always use this with @registry.rebuild_on_registers if you want the model
               to automatically support newly registered types.

            2. The Annotated type should use the registry's base class as its first argument
               (e.g., Annotated[AnimalBase, registry.DynamicResolution()])

            3. This can be used with any Pydantic field type that accepts a model, including
               nested within lists, dicts, Optional[], etc.

            4. The field will validate that the discriminator field matches one of the
               registered types and will raise a validation error if an unknown type
               is provided.

        This method is typically used in conjunction with rebuild_on_registers to create
        fully dynamic models that can handle new types as they are registered:
        ```python
        # In your base package:
        class AnimalBase(Config, ABC):
            type: str
            @abstractmethod
            def make_sound(self) -> str: ...

        registry = Registry(AnimalBase, "type")

        @registry.register
        class DogConfig(AnimalBase):
            type: Literal["dog"]
            name: str
            def make_sound(self) -> str: return "Woof!"

        @registry.rebuild_on_registers
        class ProgramConfig(Config):
            animal: Annotated[AnimalBase, registry.DynamicResolution()]

        # In a separate plugin package:
        @registry.register
        class CatConfig(AnimalBase):
            type: Literal["cat"]
            name: str
            def make_sound(self) -> str: return "Meow!"

        # Both of these will work:
        ProgramConfig(animal=DogConfig(type="dog", name="Rover"))
        ProgramConfig(animal=CatConfig(type="cat", name="Whiskers"))
        ```
        """

        return self

    def __get_pydantic_core_schema__(
        self,
        source_type: Any,
        handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        return handler.generate_schema(self.type_hint())


def _extract_registry_from_annotation(annotation: Any):
    if not isinstance(annotation, Registry):
        return None
    return annotation


def _recursively_find_registry_annotations(typ: Any) -> list[Registry]:
    """Recursively find all registry annotations in a type, including in nested types.

    For example, this will find registry annotations in:
    - Annotated[T, registry]
    - list[Annotated[T, registry]]
    - Dict[str, Annotated[T, registry]]
    - etc.

    Args:
        typ: The type to check

    Returns:
        List of found registry instances
    """
    registries = []

    # Handle Annotated types
    if typing.get_origin(typ) is typing.Annotated:
        # Check annotations
        (base_type, *annotations) = typing.get_args(typ)
        for annotation in annotations:
            if (registry := _extract_registry_from_annotation(annotation)) is not None:
                registries.append(registry)

        # Recurse into base type
        registries.extend(_recursively_find_registry_annotations(base_type))

    # Handle generic types (list, dict, etc)
    elif origin := typing.get_origin(typ):
        for arg in typing.get_args(typ):
            registries.extend(_recursively_find_registry_annotations(arg))

    return registries


def extract_registries_from_field_info(field: FieldInfo):
    """Extract all registries from a Pydantic FieldInfo object.

    This function will extract all registries from the field's type and any nested types.

    Args:
        field: The field to extract registries from

    Returns:
        List of found registry instances
    """
    registries: list[Registry] = []

    # Pydantic parses the first level of annotations, so we need to use those.
    registries.extend(
        registry
        for metadata in field.metadata
        if (registry := _extract_registry_from_annotation(metadata)) is not None
    )

    # Recursively find registry annotations
    registries.extend(_recursively_find_registry_annotations(field.annotation))

    return registries
